/*
 * video.c: A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id: video-dfb.c,v 1.2 2004/05/19 19:25:58 stefan Exp $
 */

#include <sys/mman.h>
#include <sys/ioctl.h>
#include <fcntl.h> 
#include <vdr/plugin.h>
#include "video-dfb.h"
#include "utils.h"

#define COLORKEY 17,8,79

// --- cDFBVideoOut -------------------------------------------------
IDirectFB *DFBWrapper;

static DFBEnumerationResult EnumCallBack( unsigned int id, DFBDisplayLayerDescription desc, void *data )
{
    IDirectFBDisplayLayer **layer = (IDirectFBDisplayLayer **)data;

    printf("Layer %d %s",id, desc.name);
    
    printf("  Type: ");
    if (desc.type & DLTF_BACKGROUND) printf( "background " );
    if (desc.type & DLTF_GRAPHICS) printf( "graphics " );
    if (desc.type & DLTF_STILL_PICTURE) printf( "picture " );
    if (desc.type & DLTF_VIDEO) printf( "video " );
    printf( "\n" );
    
    printf( "  Caps: " );
    if (desc.caps & DLCAPS_ALPHACHANNEL) printf( "alphachannel " );
    if (desc.caps & DLCAPS_BRIGHTNESS) printf( "brightness " );
    if (desc.caps & DLCAPS_CONTRAST) printf( "contrast " );
    if (desc.caps & DLCAPS_DEINTERLACING) printf( "deinterlacing " );
    if (desc.caps & DLCAPS_DST_COLORKEY) printf( "dst_colorkey " );
    if (desc.caps & DLCAPS_FLICKER_FILTERING) printf( "flicker_filtering " );
    if (desc.caps & DLCAPS_HUE) printf( "hue " );
    if (desc.caps & DLCAPS_LEVELS) printf( "levels " );
    if (desc.caps & DLCAPS_OPACITY) printf( "opacity " );
    if (desc.caps & DLCAPS_SATURATION) printf( "saturation " );
    if (desc.caps & DLCAPS_SCREEN_LOCATION) printf( "screen_location " );
    if (desc.caps & DLCAPS_SRC_COLORKEY) printf( "src_colorkey " );
    if (desc.caps & DLCAPS_SURFACE) printf( "surface " );
    printf("\n");

    /* We take the first layer not being the primary */
    if (id != DLID_PRIMARY && desc.caps & DLCAPS_SURFACE) {
	printf("  This is our videoLayer\n");
	try {
    	    *layer = DFBWrapper->GetDisplayLayer(id);
	} catch (DFBException *ex) {
	    printf( "Caught: %s", ex);
	    return DFENUM_CANCEL;
	}
    }

    return DFENUM_OK;
}

// List Video Modes
static DFBEnumerationResult EnumVideoModeCallback(int x, int y, int bpp, void *data)
{
    printf("%dx%d@%d ",x,y,bpp);
    return DFENUM_OK;
}

cDFBVideoOut::cDFBVideoOut()
{
    DFBSurfaceDescription dsc;
    DFBDisplayLayerDescription   desc;
    DFBCardCapabilities caps;

    printf("[dfb] init\n");
    DirectFB::Init();
    dfb = DirectFB::Create();
    dfb->SetCooperativeLevel(DFSCL_EXCLUSIVE);

    dfb->GetCardCapabilities(&caps);

    printf("[dfb] RAM: %d bytes\n",caps.video_memory);

    printf("[dfb] Accellerated Functions: ");
    if (caps.acceleration_mask & DFXL_NONE) printf("none ");
    if (caps.acceleration_mask & DFXL_FILLRECTANGLE) printf("FillRectange ");
    if (caps.acceleration_mask & DFXL_DRAWRECTANGLE) printf("DrawRectange ");
    if (caps.acceleration_mask & DFXL_DRAWLINE) printf("DrawLine ");
    if (caps.acceleration_mask & DFXL_FILLTRIANGLE) printf("FillTriangle ");
    if (caps.acceleration_mask & DFXL_BLIT) printf("Blit ");
    if (caps.acceleration_mask & DFXL_STRETCHBLIT) printf("StretchBlit ");
    if (caps.acceleration_mask & DFXL_ALL) printf("All ");
    printf("\n");
    
    printf("[dfb] Drawing Flags: ");
    if (caps.drawing_flags == DSDRAW_NOFX ) printf("none ");
    if (caps.drawing_flags & DSDRAW_BLEND ) printf("Blend ");
    if (caps.drawing_flags & DSDRAW_DST_COLORKEY ) printf("Colorkeying ");
    if (caps.drawing_flags & DSDRAW_SRC_PREMULTIPLY ) printf("Src.premultiply ");
    if (caps.drawing_flags & DSDRAW_DST_PREMULTIPLY ) printf("Dst.premultiply ");
    if (caps.drawing_flags & DSDRAW_DEMULTIPLY ) printf("Demultiply ");
    if (caps.drawing_flags & DSDRAW_XOR ) printf("Xor ");
    printf("\n");

    printf("[dfb] Surface Blitting Flags: ");
    if (caps.blitting_flags == DSBLIT_NOFX ) printf("none ");
    if (caps.blitting_flags & DSBLIT_BLEND_ALPHACHANNEL ) printf("BlendAlpha ");
    if (caps.blitting_flags & DSBLIT_BLEND_COLORALPHA ) printf("BlendColorAlpha ");
    if (caps.blitting_flags & DSBLIT_COLORIZE ) printf("Colorize ");
    if (caps.blitting_flags & DSBLIT_SRC_COLORKEY ) printf("SrcColorkey ");
    if (caps.blitting_flags & DSBLIT_DST_COLORKEY ) printf("DstColorkey ");
    if (caps.blitting_flags & DSBLIT_SRC_PREMULTIPLY ) printf("SrcPremultiply ");
    if (caps.blitting_flags & DSBLIT_DST_PREMULTIPLY ) printf("DstPremultiply ");
    if (caps.blitting_flags & DSBLIT_DEMULTIPLY ) printf("Demultiply ");
    if (caps.blitting_flags & DSBLIT_DEINTERLACE ) printf("Deinterlace ");
    printf("\n");
    
    printf("[dfb] Supported video Modes are: ");
    dfb->EnumVideoModes(EnumVideoModeCallback, NULL);
    printf("\n");
    
    Xres = 768;
    Yres = 576;
    Bpp  = 32;

    try {
	printf("[dfb] Setting  Videomode to 768x576-32\n");
	dfb->SetVideoMode(Xres,Yres,Bpp); // 768 x 576 @ 32bpp is the best choice for TV-Out
    } catch (DFBException *ex) {
	try {
	    printf("[dfb] This didn't work. Trying Videomode 1024x768-32\n");
	    Xres = 1024;
	    Yres = 768;
	    Bpp  = 32;
	    dfb->SetVideoMode(Xres,Yres,Bpp); // evtl hier andere Modes probieren
	} catch (DFBException *ex) {
	    printf("[dfb] This Vidmode doen't work, too. Exiting...\n");
	    exit(1);
	}
    }

    osdLayer=dfb->GetDisplayLayer(DLID_PRIMARY); // erster Layer. Fuer OSD

    dsc.flags = DSDESC_CAPS;
    DFB_ADD_SURFACE_CAPS(dsc.caps, DSCAPS_PRIMARY);
    DFB_ADD_SURFACE_CAPS(dsc.caps, DSCAPS_FLIPPING); 

    osdSurface=dfb->CreateSurface(dsc);
    osdSurface->Clear(COLORKEY,0); //clear and
    osdSurface->Flip(); // Flip the field
    osdSurface->Clear(COLORKEY,0); //clear and
    osdSurface->Flip(); // Flip the field

    DFBWrapper=dfb;
    printf("[dfb] Enumeratig display Layers\n");
    dfb->EnumDisplayLayers(EnumCallBack, &videoLayer);

    videoSurface=videoLayer->GetSurface();
    videoSurface->Clear(COLORKEY,0); //clear and
    videoSurface->Flip(); // Flip the field
    videoSurface->Clear(COLORKEY,0); //clear and
    videoSurface->Flip(); // Flip the field

    if(!videoLayer) {
        printf("could not find suitable videolayer\n");
        exit(1);
    } else {
	desc = osdLayer->GetDescription();
	printf("Using this layer for OSD: %s\n", desc.name);
	printf("Size: %d x %d \n",osdSurface->GetWidth(), osdSurface->GetHeight());
	DFBSurfaceCapabilities scaps;
	scaps = osdSurface->GetCapabilities();
	printf("Caps:\n");
	if (scaps & DSCAPS_NONE) printf(" - none\n");
	if (scaps & DSCAPS_PRIMARY) printf(" - primary\n");
	if (scaps & DSCAPS_SYSTEMONLY) printf(" - systemonly\n");
	if (scaps & DSCAPS_VIDEOONLY) printf(" - videoonly\n");
	if (scaps & DSCAPS_FLIPPING) printf(" - flipping\n");
	if (scaps & DSCAPS_SUBSURFACE) printf(" - subsurface\n");
	if (scaps & DSCAPS_INTERLACED) printf(" - interlaced\n");
	if (scaps & DSCAPS_SEPARATED) printf(" - separated\n");
	if (scaps & DSCAPS_STATIC_ALLOC) printf(" - static alloc\n");
	if (scaps & DSCAPS_TRIPLE) printf(" - triple buffered\n");
	scaps = videoSurface->GetCapabilities();
	printf("VideoCaps:\n");
	if (scaps & DSCAPS_NONE) printf(" - none\n");
	if (scaps & DSCAPS_PRIMARY) printf(" - primary\n");
	if (scaps & DSCAPS_SYSTEMONLY) printf(" - systemonly\n");
	if (scaps & DSCAPS_VIDEOONLY) printf(" - videoonly\n");
	if (scaps & DSCAPS_FLIPPING) printf(" - flipping\n");
	if (scaps & DSCAPS_SUBSURFACE) printf(" - subsurface\n");
	if (scaps & DSCAPS_INTERLACED) printf(" - interlaced\n");
	if (scaps & DSCAPS_SEPARATED) printf(" - separated\n");
	if (scaps & DSCAPS_STATIC_ALLOC) printf(" - static alloc\n");
	if (scaps & DSCAPS_TRIPLE) printf(" - triple buffered\n");
    }
    
    // evtl wird die eine oder andere Funktion nicht von der Graka unterstuetzt
    printf("[dfb] Configuring CooperativeLevel for Overlay\n");
    videoLayer->SetCooperativeLevel(DLSCL_ADMINISTRATIVE);
    printf("[dfb] Configuring ColorKeying\n");
    videoLayer->SetDstColorKey(COLORKEY);
    printf("[dfb] Configuring CooperativeLevel for OSD\n");
    osdLayer->SetCooperativeLevel(DLSCL_ADMINISTRATIVE);
    printf("... configured\n");
    // Wenn wir hier sind. haben wir den grten Teil schon geschafft
    desc = osdLayer->GetDescription();
    printf("[dfb] Using this layer for OSD: %s\n", desc.name);
    desc = videoLayer->GetDescription();
    printf("[dfb] Using this layer for Video out: %s\n", desc.name);

}

void cDFBVideoOut::SetParams(int newXres, int newYres)
{
    DFBDisplayLayerConfig dlc;
    DFBDisplayLayerDescription desc;

    if (videoLayer)
    {
	if (!videoSurface || (ovlXres !=newXres) || (ovlYres != newYres))
	{
	    printf("[dfb] (re)configuring Videolayer to %d x %d \n", newXres,newYres);

	    dlc.flags   = (DFBDisplayLayerConfigFlags)(DLCONF_WIDTH | DLCONF_HEIGHT | DLCONF_PIXELFORMAT | DLCONF_OPTIONS);
	    dlc.width   = newXres;
	    dlc.height  = newYres;
	    dlc.options = (DFBDisplayLayerOptions)(DLOP_DST_COLORKEY | DLOP_DEINTERLACING ); // maybe your graphic card does not support this

	    dlc.pixelformat = DSPF_I420;
//	    dlc.pixelformat = DSPF_YUY2;

	    dlc.options = (DFBDisplayLayerOptions)( DLOP_ALPHACHANNEL );
//	    dlc.options = (DFBDisplayLayerOptions)( DLOP_DST_COLORKEY | DLOP_DEINTERLACING ); // maybe your graphic card does not support this
//	    dlc.options = (DFBDisplayLayerOptions)( DLOP_NONE ); // maybe you should try this

	    pixelformat = dlc.pixelformat;

	    try {
		videoLayer->SetConfiguration(dlc);
	    } catch (DFBException *ex) {
		printf("Caught: %s", ex);
		exit(1);
	    }

	    desc = videoLayer->GetDescription();

	    if (desc.caps & DLCAPS_SCREEN_LOCATION) {
		videoLayer->SetScreenLocation(0.0f, 0.0f, 0.9999f, 1.0f); // my Matrox doesn't wor when setting to 0.0 0.0 1.0 1.0
	    } else {
		printf("Can't configure ScreenLocation. Hope it is Fullscreen\n");
	    }

	    videoSurface=videoLayer->GetSurface();
	    ovlXres = newXres;
	    ovlYres = newYres;
	    printf("[dfb] (re)configured\n");
	}
    } else {
	printf("No Videolayer avaiable. Exiting...\n");
	exit(1);
    }
}

void cDFBVideoOut::Pause(void) 
{
}

#if VDRVERSNUM >= 10307

void cDFBVideoOut::Refresh(cBitmap *Bitmap)
{
    const tIndex *adr;
    tIndex *buf;
    tColor c;
    int pitch;
    uint8_t *dst;

    osdSurface->Lock(DSLF_WRITE, (void **)&dst, &pitch) ;

    for (int y = 0; y < Bitmap->Height(); y++)
    {
	buf = (tIndex *) dst + pitch * ( OSDyOfs + y ) + OSDxOfs * 4;

	for (int x = 0; x < Bitmap->Width(); x++)
	{
	    adr = Bitmap->Data(x, y);
	    c = Bitmap->Color(*adr);

	    buf[3] = (c >> 24) & 255; //Alpha
	    buf[2] = (c >> 16) & 255; //Red
	    buf[1] = (c >> 8) & 255;  //Green
	    buf[0] = c & 255;         //Blue

	    buf+=4;
	}
    }

    osdSurface->Unlock();
    osdSurface->Flip();
}

#else

void cDFBVideoOut::Refresh()
{
    int pitch;
    uint8_t *dst;

    osdSurface->Lock(DSLF_WRITE, (void **)&dst, &pitch) ;
    for (int i = 0; i < MAXNUMWINDOWS; i++) {
	    if (layer[i] && layer[i]->visible) layer[i]->Draw(dst, pitch, NULL);
    }
    osdSurface->Unlock();
    osdSurface->Flip();
}

#endif

void cDFBVideoOut::YUV(uint8_t *Py, uint8_t *Pu, uint8_t *Pv, int Width, int Height, int Ystride, int UVstride)
{
    uint8_t *dst;
    int pitch;
    int hi;

    SetParams(Width, Height);
    
//  printf("[dfb] draw frame (%d x %d) Y: %d UV: %d\n", Width, Height, Ystride, UVstride);

    videoSurface->Lock(DSLF_WRITE, (void **)&dst, &pitch);

    if (pixelformat == DSPF_I420)
    {
	for(hi=0; hi < Height; hi++){
	    memcpy(dst, Py, Width);
	    Py  += Ystride;
            dst += pitch;
	}

        for(hi=0; hi < Height/2; hi++) {
            memcpy(dst, Pu, Width/2);
            Pu  += UVstride;
            dst += pitch / 2;
        }

        for(hi=0; hi < Height/2; hi++) {
            memcpy(dst, Pv, Width/2);
	    Pv  += UVstride;
            dst += pitch / 2;
	}

	videoSurface->Unlock();

    } else if (pixelformat == DSPF_YUY2) {

#ifndef USE_MMX
	/* reference implementation */
	int p = pitch - Width*2;
	for (int i=0; i<Height; i++) {
	    for (int j =0; j < Width/2; j++) {
		*dst = *(Py + (j*2) + i * Ystride);
		dst +=1;
		*dst = *(Pu + j + (i/2) * UVstride);
		dst +=1;
		*dst = *(Py + (j*2)+1 + i * Ystride);
		dst +=1;
		*dst = *(Pv + j + (i/2) * UVstride);
		dst +=1;
	    }
	    dst +=p;
	}
# else
	/* deltas */
	for (int i=0; i<Height; i++) {
	    uint8_t *pu, *pv, *py, *srfc;
	    pu = Pu;
	    pv = Pv;
	    py = Py;
	    srfc = dst;
	    for (int j =0; j < Width/8; j++) {
		movd_m2r(*pu, mm1);		// mm1 = 00 00 00 00 U3 U2 U1 U0
		movd_m2r(*pv, mm2);		// mm2 = 00 00 00 00 V3 V2 V1 V0
		movq_m2r(*py, mm0);		// mm0 = Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0
		punpcklbw_r2r(mm2, mm1);	// mm1 = V3 U3 V2 U2 V1 U1 V0 U0
		movq_r2r(mm0,mm3);		// mm3 = Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0
		movq_r2r(mm1,mm4);		// mm4 = V3 U3 V2 U2 V1 U1 V0 U0
		punpcklbw_r2r(mm1, mm0);	// mm0 = V1 Y3 U1 Y2 V0 Y1 U0 Y0
		punpckhbw_r2r(mm4, mm3);	// mm3 = V3 Y7 U3 Y6 V2 Y5 U2 Y4
	    
		movntq(mm0,*srfc);	// Am Meisten brauchen die Speicherzugriffe
		srfc+=8;
		py+=8;
		pu+=4;
		pv+=4;
		movntq(mm3,*srfc);	// wenn movntq nicht geht, dann movq verwenden
		srfc+=8;
	    }
	    Py += Ystride;;
	    if (i % 2 == 1) {
		Pu += UVstride;
		Pv += UVstride;
	    }
	    dst += pitch;
	}
# endif
	videoSurface->Unlock();
    }
/*
    try {
	videoSurface->Flip();
    } catch (DFBException *ex){
	printf("Flip failed\n");
    }
*/
}

void cDFBVideoOut::CloseOSD()
{
    osdSurface->Clear(COLORKEY,0); //clear and
    osdSurface->Flip(); // Flip the field
    osdSurface->Clear(COLORKEY,0); //clear and
    osdSurface->Flip(); // Flip the field
}

cDFBVideoOut::~cDFBVideoOut()
{
    printf("Releasing DFB\n");
    if (dfb) dfb->Release();
}
